#include "spawn.h"

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"


//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}


//////// Request queue ////////

void botInfo::add_request(Player *p)
{
	InfoRequest *info = new InfoRequest;

	strncpy(info->name, p->name, 20);
	info->player_id = p->ident;
	info->request_time = GetTickCount();
	info->alias = NULL;

	queue.append(info);

	if (queue.total == 1)
	{
		sendPrivate(p, "*info");
	}
}

Player *botInfo::get_request()
{
	if (queue.total == 0 || queue.head == 0)
		return NULL;

	InfoRequest *info = queue.head->item;

	_listnode <Player> *parse = playerlist->head;

	while (parse)
	{
		Player *p = parse->item;

		if (p->ident == info->player_id)
		{
			if (CMPSTR(p->name, info->name))
			{
				return p;
			}
		}

		parse = parse->next;
	}

	return NULL;
}

void botInfo::dequeue()
{
	if (queue.head)
		queue.kill(queue.head);
}

void botInfo::request_next()
{
	if (queue.total == 0)
		return;

	Player *p = get_request();

	if (!p)
	{
		dequeue();
		request_next();
		return;
	}

	sendPrivate(p, "*info");
	countdown[1] = 5;
}

void botInfo::CheckZip(Player *p, alias_name *alias)
{
	zip_info *zip = MatchZip(alias);

	if (zip)
	{
		if (zip->b)
		{
			String s;
			s += "*warn You have been banned from this arena for ";
			s += zip->minutes;
			s += " minutes. ";
			s += zip->minutes - (ZipElapsed(zip) / 60);
			s += " minutes remaining on your ban.";
			sendPrivate(p, s);
			sendPrivate(p, "*kill");
			if (send_to_channel)
			{
				s = "1;Zip-banned player in arena ";
				s += arena;
				s += ": ";
				s += p->name;
				sendChannel(s);
			}
		}
		else
		{
			ZipRequest *rq	= new ZipRequest;
			rq->request_time= GetTickCount();
			rq->p			= p;
			rq->c			= false;
			rq->s			= false;
			rq->s_ok		= false;
			rq->c_ok		= false;

			if (zip->c)
			{
				if (zip->s)
				{
					String s;
					s += "*warn You have been spec-locked and silenced in this arena for ";
					s += zip->minutes;
					s += " minutes. ";
					s += zip->minutes - (ZipElapsed(zip) / 60);
					s += " minutes remaining.";
					sendPrivate(p, s);
					sendPrivate(p, "*spec");
					sendPrivate(p, "*shutup");
					if (send_to_channel)
					{
						s = "1;Zip-silenced&speclocked player in arena ";
						s += arena;
						s += ": ";
						s += p->name;
						sendChannel(s);
					}

					rq->c = true;
					rq->s = true;
				}
				else
				{
					String s;
					s += "*warn You have been spec-locked in this arena for ";
					s += zip->minutes;
					s += " minutes. ";
					s += zip->minutes - (ZipElapsed(zip) / 60);
					s += " minutes remaining.";
					sendPrivate(p, s);
					sendPrivate(p, "*spec");
					if (send_to_channel)
					{
						s = "1;Zip-speclocked player in arena ";
						s += arena;
						s += ": ";
						s += p->name;
						sendChannel(s);
					}

					rq->c = true;
					rq->s_ok = true;
				}
			}
			else
			{
				String s;
				s += "*warn You have been silenced in this arena for ";
				s += zip->minutes;
				s += " minutes. ";
				s += zip->minutes - (ZipElapsed(zip) / 60);
				s += " minutes remaining.";
				sendPrivate(p, s);
				sendPrivate(p, "*shutup");
				if (send_to_channel)
				{
					s = "1;Zip-silenced player in arena ";
					s += arena;
					s += ": ";
					s += p->name;
					sendChannel(s);
				}

				rq->s = true;
				rq->c_ok = true;
			}

			zip_requests.append(rq);
		}
	}
}

bool CMPEND(char *text, char *end)
{
	int tend = STRLEN(text);
	int eend = STRLEN(end);

	if (tend < eend)
		return false;

	int start = tend - eend;

	for (int i = start; i < tend; ++i)
		if (text[i] != end[i - start])
			return false;

	return true;
}


//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
			for (int i = 0; i < 4; ++i)
				--countdown[i];

			if (countdown[0] == 0)
			{
				if (ListChanged)
				{
					if (save_aliases())
					{
						tell(makeEcho("Saved aliases."));
						ListChanged = false;
						countdown[0] = 60 * 5;
					}
					else
					{
						tell(makeEcho("Unable to save aliases."));
						countdown[0] = 30;
					}
				}
			}

			if (countdown[1] == 0)
			{
				if (info_valid)
				{
					if (!queue.head)
					{
					}
					else
					{
						if (queue.head->item)
						{
							alias_name *alias = queue.head->item->alias;

							work.timediff = 0;

							if (!alias)
							{
								alias = AddAlias(work);

								UpdateHash(*alias);
							}

							AddData(alias, work);
							UpdateSeen(*alias);
						}

						dequeue();
						request_next();

						info_valid = false;
					}

					info_valid = false;
				}
				else
				{
					queue.clear();
				}
			}

			if (countdown[2] == 0)
			{
				countdown[2] = 10;

				_listnode <zip_info> *parse = ZipList.head;
				_listnode <zip_info> *kp;

				while (parse)
				{
					kp = parse;
					zip_info *zip = kp->item;
					parse = kp->next;

					if (zip->minutes - (ZipElapsed(zip) / 60) <= 0)
					{
						_listnode <Player> *pparse = playerlist->head;

						while (pparse)
						{
							Player *pp = pparse->item;

							if (CMPSTR(zip->alias->name, pp->name))
							{
								String s = "*warn Zip time up.";

								// Make sure that the player has the proper status
								ZipRequest *rq = new ZipRequest;
								rq->request_time= GetTickCount();
								rq->p = pp;
								rq->c = false;
								rq->s = false;
								rq->c_ok = !zip->c;
								rq->s_ok = !zip->s;
								zip_requests.append(rq);

								if (zip->c)
								{
									s += " You may now enter the arena.";
									sendPrivate(pp, "*spec");
								}

								if (zip->s)
								{
									s += " You may now speak.";
									sendPrivate(pp, "*shutup");
								}

								sendPrivate(pp, s);

								break;
							}

							pparse = pparse->next;
						}

						ZipList.kill(kp);
					}
				}
			}

			if (countdown[3] == 0)
			{
				int time = GetTickCount();

				// Clean up the info request queue
				{
					_listnode<InfoRequest> *parse = queue.head;

					while (parse)
					{
						if (time - parse->item->request_time > 2500)
						{
							queue.clear();
							break;
						}

						parse = parse->next;
					}
				}

				// Clean up the zip request queue
				{
					_listnode<ZipRequest> *parse = zip_requests.head;

					while (parse)
					{
						if (time - parse->item->request_time > 2500)
						{
							zip_requests.clear();
							break;
						}

						parse = parse->next;
					}
				}

				countdown[3] = 2;
			}
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}
			biller_online = *(bool*)&event.p[2];

			if (!biller_online) break;

			_listnode <Player> *parse = playerlist->head;

			while (parse)
			{
				Player *p = parse->item;
				parse = parse->next;

				if (p == me) continue;

				add_request(p);
			}
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
			killTags();
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
				p[i] = (Player*)event.p[i];
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_BallMove:
		{
			PBall *ball = (PBall*)event.p[0];
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];

			if (!biller_online) break;

			add_request(p);
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];
		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];

			killTags(p);

			alias_name *alias = FindSnapshot(p->name);

			if (alias)
				UpdateSeen(*alias);

			_listnode <ZipRequest> *parse = zip_requests.head;

			while (parse)
			{
				ZipRequest *rq = parse->item;

				if (rq->p == p)
				{
					rq->p = NULL;
				}

				parse = parse->next;
			}
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				{
					if (!biller_online) break;

					if (sound != SND_None)
						break;

					if (CMPSTR("Player locked in spectator mode", msg))
					{	// *spec ON
						if (!zip_requests.head) break;
						ZipRequest *rq = zip_requests.head->item;
						if (rq->c_ok) break;

						if (!rq->c)
						if (rq->p)
						{
							sendPrivate(rq->p, "*spec");

							ZipRequest *nrq = new ZipRequest;
							nrq->request_time = GetTickCount();
							nrq->p = NULL;
							nrq->c = false;
							nrq->s = false;
							nrq->c_ok = false;
							nrq->s_ok = true;
							zip_requests.append(nrq);
						}

						rq->c_ok = true;

						if (rq->s_ok)
							zip_requests.kill(zip_requests.head);
					}
					else if (CMPSTR("Player free to enter arena", msg))
					{	// *spec OFF
						if (!zip_requests.head) break;
						ZipRequest *rq = zip_requests.head->item;
						if (rq->c_ok) break;

						if (rq->c)
						if (rq->p)
						{
							sendPrivate(rq->p, "*spec");

							ZipRequest *nrq = new ZipRequest;
							nrq->request_time = GetTickCount();
							nrq->p = NULL;
							nrq->c = true;
							nrq->s = false;
							nrq->c_ok = false;
							nrq->s_ok = true;
							zip_requests.append(nrq);
						}

						rq->c_ok = true;

						if (rq->s_ok)
							zip_requests.kill(zip_requests.head);
					}
					else if (CMPEND(msg, " has been silenced"))
					{	// *shutup ON
						if (!zip_requests.head) break;
						ZipRequest *rq = zip_requests.head->item;
						if (rq->s_ok) break;

						if (!rq->s)
						if (rq->p)
						{
							sendPrivate(rq->p, "*shutup");

							ZipRequest *nrq = new ZipRequest;
							nrq->request_time = GetTickCount();
							nrq->p = NULL;
							nrq->c = false;
							nrq->s = false;
							nrq->c_ok = true;
							nrq->s_ok = false;
							zip_requests.append(nrq);
						}

						rq->s_ok = true;

						if (rq->c_ok)
							zip_requests.kill(zip_requests.head);
					}
					else if (CMPEND(msg, " can now speak"))
					{	// *shutup OFF
						if (!zip_requests.head) break;
						ZipRequest *rq = zip_requests.head->item;
						if (rq->s_ok) break;

						if (rq->s)
						if (rq->p)
						{
							sendPrivate(rq->p, "*shutup");

							ZipRequest *nrq = new ZipRequest;
							nrq->request_time = GetTickCount();
							nrq->p = NULL;
							nrq->c = false;
							nrq->s = true;
							nrq->c_ok = true;
							nrq->s_ok = false;
							zip_requests.append(nrq);
						}

						rq->s_ok = true;

						if (rq->c_ok)
							zip_requests.kill(zip_requests.head);
					}
					else if (parse_info(msg, work))
					{
						if (queue.total == 0)
							break;

						alias_name *alias;

						p = get_request();

						if (p)
						{
							alias = FindSnapshot(p->name);
						}
						else
						{
							alias = FindSnapshot(work.name);
						}

						if (alias)
						if (old_info(alias, work))
						{
							if (p)
							{	// do zipping
								CheckZip(p, alias);
							}

							dequeue();
							request_next();

							break;
						}

						if (!p)
						{
							work.timediff = 0;

							if (!alias)
							{
								alias = AddAlias(work);

								UpdateHash(*alias);
							}

							AddData(alias, work);
							UpdateSeen(*alias);

							dequeue();
							request_next();

							break;
						}

						queue.head->item->alias = alias;

						info_valid = true;
						sendPrivate(p, "*tinfo");
						countdown[1] = 5;
					}
					else if (parse_tinfo(msg, work) && (info_valid))
					{
						if (queue.total == 0)
						{
							break;
						}

						alias_name *alias = queue.head->item->alias;

						if (!alias)
						{
							alias = AddAlias(work);

							UpdateHash(*alias);
						}

						AddData(alias, work);
						UpdateSeen(*alias);

						p = get_request();

						if (p)
						{
							CheckZip(p, alias);
						}

						dequeue();
						request_next();

						info_valid = false;
					}
				}
				break;
			case MSG_PublicMacro:		if (!p) break;
				{
				}
				break;
			case MSG_Public:			if (!p) break;
				{
				}
				break;
			case MSG_Team:				if (!p) break;
				{
				}
				break;
			case MSG_TeamPrivate:		if (!p) break;
				{
				}
				break;
			case MSG_Private:			if (!p) break;
				{
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major = HIWORD(*(int*)&event.p[0]);
			int minor = LOWORD(*(int*)&event.p[0]);
			callback = (CALL_COMMAND)event.p[1];
			playerlist = (CALL_PLIST)event.p[2];
			flaglist = (CALL_FLIST)event.p[3];

			if (major > CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin is out of date."));

				CONNECTION_DENIED = true;

				return;
			}
			else if ((major < CORE_MAJOR_VERSION) || (minor < CORE_MINOR_VERSION))
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin requires the latest version of MERVBot."));

				CONNECTION_DENIED = true;

				return;
			}
			else
				tell(makeEcho("DLL plugin connected."));

			if (AliasList.total == 0)
			{
				load_aliases();
			}
		}
		break;
	case EVENT_Term:
		{
			if (ListChanged)
			{
				tell(makeEcho("Saved aliases."));

				save_aliases();
			}

			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	};
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Tag emulation ////////

int botInfo::get_tag(Player *p, int index)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
			return tag->data;

		parse = parse->next;
	}

	return 0;
}

void botInfo::set_tag(Player *p, int index, int data)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
		{
			tag->data = data;
			return;
		}

		parse = parse->next;
	}

	tag = new PlayerTag;
	tag->p = p;
	tag->index = index;
	tag->data = data;
	taglist.append(tag);
}

void botInfo::killTags(Player *p)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;
		parse = parse->next;

		if (tag->p != p) continue;

		taglist.kill(tag);
	}
}

void botInfo::killTags()
{
	taglist.clear();
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s);
}
